/***************************************************************************
 *   Copyright (C) 2007 by Sergey Semionov   							   *
 *   simon@vu.spb.ru                                                       *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/
 

/***************************************************************************
 * This driver is fully based on the vec535 driver.                         *
 * Thank you Jeroen Vreeken for the great work!                            *
 *                                                                         *
 * Also would like to thank the EVS company (www.evs.ru) for the wonderful *
 * and low-cost USB WEB Cams and especially Vitaly (vitality@mail.ru) for  * 
 * helping me with driver development.                                     *
 *                                    -- Sergey Semionov                   *
 ***************************************************************************/

#include <linux/module.h>
#include <linux/init.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/pagemap.h>
#include <linux/usb.h>
#include "vec535-driver.h"

static const char version[] = "0.2"; 
static int video_nr = -1;

MODULE_DESCRIPTION("VEC535 USB Camera Driver");
MODULE_AUTHOR("Sergey Semionov (simon@vu.spb.ru)");
MODULE_LICENSE("GPL");

static int debug = 5;
module_param(debug, int, 5);
MODULE_PARM_DESC(debug,"Debug level: 0=none, 1=simple, 2=full");

static int enhance = 0;
module_param(enhance, int,0);
MODULE_PARM_DESC(debug, "Use enhancing algorithm");

static int levelAE = 88;
module_param(levelAE, int, 88);
MODULE_PARM_DESC(levelAE, "upper level for APY/APBH");

static int loopAE = 68;
module_param(loopAE, int, 68);
MODULE_PARM_DESC(loopAE, "lower level for APY/APBH");

static int gamma = 2;
module_param(gamma, int, 0);
MODULE_PARM_DESC(gamma, "Gamma level (0,1, 2, 3)");

static struct usb_device_id device_table[] = {
	{USB_DEVICE(0x04b4,0x0535)}, 
	{ }
};
MODULE_DEVICE_TABLE(usb, device_table);

/**********************************************************************
 * Memory management
 **********************************************************************/
static void* rvmalloc(unsigned long size)
{
	void *mem;
	unsigned long adr;

	size = PAGE_ALIGN(size);
	mem = (void *)vmalloc_32(size);
	if (!mem)
		return NULL;

	memset(mem, 0, size); /* Clear the ram out, no junk to the user */
	adr = (unsigned long) mem;
	while (size > 0) {
		SetPageReserved(vmalloc_to_page((void *)adr));
		adr += PAGE_SIZE;
		size -= PAGE_SIZE;
	}

	return mem;
}

static void rvfree(void *mem, unsigned long size)
{
	unsigned long adr;

	if (!mem)
		return;

	adr = (unsigned long) mem;
	while ((long) size > 0) {
		ClearPageReserved(vmalloc_to_page((void *)adr));
		adr += PAGE_SIZE;
		size -= PAGE_SIZE;
	}
	vfree(mem);
}

static void calculateSaturation(struct usb_vec535 *vec535, unsigned short colour) {
	colour = colour >> 8;

	vec535->Sx = ((colour * 256)) / 48;
	vec535->Sr = 77 - ((short) colour * 77) / 48;
	vec535->Sg = 150 - ((short) colour * 150) / 48;
	vec535->Sb = 29 - ((short) colour * 29) / 48;

}

/**********************************************************************
 * Control pipe
 **********************************************************************/
static int vec535_control(int set, struct usb_vec535 *vec535, unsigned short req,
			 unsigned short value, unsigned short index)
{
	int ret = usb_control_msg (
		vec535->udev,
		set ? usb_sndctrlpipe(vec535->udev, 0) : usb_rcvctrlpipe(vec535->udev, 0),
		req, 
		(set ? USB_DIR_OUT : USB_DIR_IN) | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		value,
		index,
		NULL,
		0,
		1000
	);

	if (ret < 0) 
		PDEBUG(3, "returned %d for 0x%X 0x%X 0x%X", ret, req, value, index);

	return ret;
}

static int _WRITE_REG(struct usb_vec535 *vec535, unsigned short Reg,
			     unsigned short Value)
{
	/* specs say that the selector (address) should go in the value field
	   and the param in index, but in the logs of the windows driver they do
	   this the other way around...
	 */
	int retvalue;
	retvalue= usb_control_msg (
		vec535->udev,
		usb_sndctrlpipe(vec535->udev, 0),
		0x02,
		USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		Value, 
		Reg,
                NULL,
                0,
                HZ
        );
	if (retvalue != 0) {
		PDEBUG(3, "_WRITE_REG fails with error %d", retvalue);
	}
	else {
		//PDEBUG(3, "_WRITE_REG REG=0x%x VALUE=%d", Reg, Value);
	}

	return retvalue;
}

static int _WRITE_PORTD(struct usb_vec535 *vec535, unsigned short Reg,
			     unsigned short Value)
{
	/* specs say that the selector (address) should go in the value field
	   and the param in index, but in the logs of the windows driver they do
	   this the other way around...
	 */
	int retvalue;
	retvalue= usb_control_msg (
		vec535->udev,
		usb_sndctrlpipe(vec535->udev, 0),
		0x06,
		USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		Value, 
		Reg,
                NULL,
                0,
                HZ
        );

	return retvalue;
}




static unsigned short _READ_REG(struct usb_vec535 *vec535, 
				        unsigned short Reg)
{
	/* For 'set' the selecetor should be in index, not sure if the spec is
	   wrong here to....
	 */
	int retvalue;
	unsigned char cp[] = {0, 0};

	retvalue = usb_control_msg(
		vec535->udev,
		usb_rcvctrlpipe(vec535->udev, 0),
		0x03/*REQUEST*/,
		USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		0x00/*VALUE*/,
		Reg/*INDEX*/,
		cp,
		2,
		HZ);

	if (retvalue == 1) {
		//PDEBUG(3, " _READ_REG REG=0x%X VALUE=%d", Reg, cp[0]);
		return cp[0];

	}else {
		//PDEBUG(3, "_READ_REG returns 0x%X with 0x%X 0x%X", retvalue, cp[0], cp[1]);
		return (cp[1] << 8) | cp[0];
	}
}
#define WRITE_REG(Reg, Value) _WRITE_REG(vec535, Reg,Value)
#define READ_REG(Reg) _READ_REG(vec535, Reg)
#define WRITE_PORTD(Reg, Value) _WRITE_PORTD(vec535, Reg, Value)

#define ISAUTOMODE(Value) ((Value < 512) ? 1 : 0)
static int vec535_get_pict(struct usb_vec535 *vec535, struct video_picture *p) {
	
	//short Value;
	//int test;

	//Get HUE
	//Value = (short) READ_REG(0x02, test);
	//PDEBUG(3, "HUE is %d, %d", Value, test);
	
	//p->hue = Value;
	return 1;
}

/*static void vec535_init_pict(struct usb_vec535 *vec535, struct video_picture *p) {
	//disable all auto

	unsigned short Value, ValueRead;

	ValueRead = READ_REG(0x13);
	Value = ValueRead & 0xF8;
	WRITE_REG(0x13, Value);
	
}*/
static int vec535_set_pict(struct usb_vec535 *vec535, struct video_picture *p)
{
	unsigned short Value, ValueRead, Value2Set;

	//vec535_init_pict(vec535, p);

	if (p->palette != VIDEO_PALETTE_RGB24)
		return 1;

	if (p->hue != vec535->hue) {
	    	Value = (vec535->hue) >> 8;
		vec535->hue = p->hue;
		//FIXME
	
		WRITE_REG(0x02, Value);
	}	
	
	ValueRead = READ_REG(0x13);

	PDEBUG(3, "autolevel = %d", ValueRead & 0x7);

	if (p->whiteness != vec535->whiteness) {

		if (ISAUTOMODE(p->whiteness)) {
			if (!ISAUTOMODE(vec535->whiteness)) {
				ValueRead = READ_REG(0x13);
				Value2Set = ValueRead | 2;
				WRITE_REG(0x13, Value2Set);
				PDEBUG(3, "Enable auto whitebalance");
			}
			
		}else { 
			if (ISAUTOMODE(vec535->whiteness)) {
				ValueRead = READ_REG(0x13);
				Value2Set = ValueRead & 0xFD;
				WRITE_REG(0x13, Value2Set);

				PDEBUG(3, "Disable auto whitebalance");
			}
		    	Value = ((vec535->whiteness) >> 8);

			WRITE_REG(0x1, Value);
		}
		vec535->whiteness = p->whiteness;
	}

	if (p->brightness!=vec535->brightness) {
		
		if (ISAUTOMODE(p->brightness)) { //Exposure
			
			if (!ISAUTOMODE(vec535->brightness)) {
				ValueRead = READ_REG(0x13);
				Value2Set = ValueRead | 1;
				WRITE_REG(0x13, Value2Set);
				PDEBUG(3, "Enable auto brightness");
			}
		}else {
			if (ISAUTOMODE(vec535->brightness)) {
				ValueRead = READ_REG(0x13);
				Value2Set = ValueRead & 0xFE;
				WRITE_REG(0x13, Value2Set);
				PDEBUG(3, "Disable auto brightness");
			}

			Value = (p->brightness) / 21; //should be in interval 0..3071

			ValueRead = READ_REG(0x04);
			ValueRead = ((ValueRead & 0xF8) | (Value & 0x07));
			WRITE_REG(0x04, ValueRead);

			Value2Set = ((Value >> 3) & 0xFF);
			WRITE_REG(0x10, Value2Set);

			ValueRead = READ_REG(0x45);
			Value2Set = ((ValueRead & 0xC0) | ((Value >> 11) & 0x3F));
			WRITE_REG(0x45, Value2Set);
		}
		vec535->brightness = p->brightness;	
	}

	if (p->contrast != vec535->contrast) {
		if (ISAUTOMODE(p->contrast)) {	//REQ_GAIN
			//enable AUTO mode
			if (!ISAUTOMODE(vec535->contrast)) {
				ValueRead = READ_REG(0x13);
				Value2Set = ValueRead | 4;
				WRITE_REG(0x13, Value2Set);
			
				PDEBUG(3, "Enable auto rgain");
			}
		}else {
			if (ISAUTOMODE(vec535->contrast)) {
				ValueRead = READ_REG(0x13);
				Value2Set = ValueRead & 0xFB;
				WRITE_REG(0x13, Value2Set);
				PDEBUG(3, "Disable auto rgain");
			}
			Value = (p->contrast) / 590; //interval 0..111


			if (Value > 0x5F) Value += 0x1F0;
			if (Value > 0x4F) Value += 0xF0;
			if (Value > 0x3F) Value += 0x70;
			if (Value > 0x2F) Value += 0x30;
			if (Value > 0x1F) Value += 0x10;
		
			WRITE_REG(0x00, (Value & 0xFF));

			ValueRead = READ_REG(0x45);
			Value2Set = ((ValueRead & 0x3F) | ((Value >> 2) & 0xC0));
			WRITE_REG(0x45, Value2Set);
		}
		vec535->contrast = p->contrast;

	}
	
	if (p->colour != vec535->colour) {
		vec535->colour = p->colour;
		calculateSaturation(vec535, p->colour);
	}
	/*
	if (p->brightness!=vec535->brightness) {
		vec535_set_exposure(vec535, p->brightness);
	}
	if (p->whiteness != vec535->whiteness) {
	    CurrValue = vec535->whiteness;
		vec535->whiteness = p->whiteness;
  
		Status = READ_REG(0x13, CurrValue, Length);
		CurrValue = CurrValue & 0xFD;
		Status = WRITE_REG(0x13, CurrValue);
		if (Value<0) CurrValue = 0x80 + (short)Value;
		  else CurrValue = 0x80 | (short)Value;
		Status = WRITE_REG(0x01,CurrValue);
		
		PDEBUG(1, "SET new whiteness %d", p->whiteness);
		
	}
	*/
	return 0;
}

static void vec535_setMode(struct usb_vec535 *vec535, int mode) {
	switch (mode) {
		case MODE_864x600:      //on this mode packets with leng 256 comes on bayer offset 518144 to be exactul 518400
					vec535_control(1, vec535, 0x02, 0x20, 0x12);
					vec535_control(1, vec535, 0x02, 0x11, 0x17);
					vec535_control(1, vec535, 0x02, 0x47, 0x18);
					vec535_control(1, vec535, 0x02, 0x01, 0x19);
					vec535_control(1, vec535, 0x02, 0x4c, 0x1A);
					vec535_control(1, vec535, 0x02, 0x36, 0x32);
					vec535_control(1, vec535, 0x02, 0x45, 0x03);

					vec535_control(1, vec535, 0x02, 0x40, 0x11); //was 0x41
					vec535->cwidth  = 864;
					vec535->cheight = 600; 
					
					PDEBUG(3, "MODE_864x600");
					break;
		case MODE_1280x960:	
					vec535_control(1, vec535, 0x02, 0x40, 0x12);
					vec535_control(1, vec535, 0x02, 0x11, 0x17);
					vec535_control(1, vec535, 0x02, 0x61, 0x18);
					vec535_control(1, vec535, 0x02, 0x01, 0x19);
					vec535_control(1, vec535, 0x02, 0x79, 0x1A);
					vec535_control(1, vec535, 0x02, 0x36, 0x32);
					vec535_control(1, vec535, 0x02, 0x45, 0x03);

					vec535_control(1, vec535, 0x02, 0x40, 0x11);
					vec535->cwidth  = 1280;
					vec535->cheight = 960; 
					PDEBUG(3, "MODE_1280x960");
					
					break;
		case MODE_2136x1602:	
					vec535_control(1, vec535, 0x02, 0x00, 0x12);
					vec535_control(1, vec535, 0x02, 0x11, 0x17);
					vec535_control(1, vec535, 0x02, 0x97, 0x18);
					vec535_control(1, vec535, 0x02, 0x02, 0x19);
					vec535_control(1, vec535, 0x02, 0xCA, 0x1A);
					vec535_control(1, vec535, 0x02, 0x16, 0x32);
					vec535_control(1, vec535, 0x02, 0x44, 0x03);

					vec535_control(1, vec535, 0x02, 0x41, 0x11);
					vec535->cwidth  = 2136;
					vec535->cheight = 1602; 
					PDEBUG(3, "MODE_2136x1602");
					
					break;
		case MODE_2592x1944:	
					//FIXME: need to add
					vec535_control(1, vec535,0x02, 0x04, 0x12);
					vec535_control(1, vec535,0x02, 0x10, 0x17);
					vec535_control(1, vec535,0x02, 0xB2, 0x18);
					vec535_control(1, vec535,0x02, 0x00, 0x19);
					vec535_control(1, vec535,0x02, 0xF3, 0x1A);
					vec535_control(1, vec535,0x02, 0x36, 0x32);
					vec535_control(1, vec535,0x02, 0x4A, 0x03);
					
					vec535_control(1, vec535, 0x02, 0x42, 0x11); //was 0x41
					vec535->cwidth  = 2592;
					vec535->cheight = 1944; 
					PDEBUG(3, "MODE_2592x1944");
					break;
	
	}
	msleep(1000);
	
}
#define DEFAULT_SATURATION		96
static int vec535_init(struct usb_vec535 *vec535) {
	unsigned short portDvalue = 0;
	vec535_control(1, vec535, 0x0A, 0x01,0x0F);
	vec535_control(1, vec535, 0x0A, 0x03,0x0F);
	vec535_control(1, vec535, 0x0A, 0x01,0x0F);
	vec535_control(1, vec535, 0x0A, 0x00,0x60);
	vec535_control(1, vec535, 0x0A, 0x01,0x02);
	
	//wait for 2ms
	msleep(200);
	vec535_control(1, vec535, 0x0A, 0x15,0x0F); 

	if (levelAE) {
		WRITE_REG(0x24, levelAE);
	}
	if (loopAE) {
		WRITE_REG(0x25, loopAE);
	}
	if (gamma) {
		portDvalue |= (gamma << 4);
		WRITE_PORTD(0x06, portDvalue);
	}	
	calculateSaturation(vec535, DEFAULT_SATURATION << 8);
	return 0;
}

/**********************************************************************
 * call back function.
 * It's called when new USB packet is arrived. The packet data is stored
 * in scratch cycle buffer in order to be decoded later
 **********************************************************************/
static void vec535_video_irq(struct urb *urb)
{
	struct usb_vec535 *vec535 = urb->context;
	int length = urb->actual_length;

	if (length == 0) {
		PDEBUG(1, "irq with packet length = %d", length);
	}

	//PDEBUG(1, "irq with packet length = %d", length);
	if (!vec535->streaming)
		return;

	if (!vec535->udev) {
		PDEBUG(3, "ohoh: device vapourished");
		return;
	}

		

	//if (length == 0) {
		/* Incoming NULL packet informs about (0,0) position of the frame*/
	//	vec535->bayeroffset=0; //FIXME!!!
	//}
	if ((length != 512)) {
		switch (vec535->cwidth) {
			case	864:
				 	if (length == 256) 
						vec535->bayeroffset = 518144;	//864 * 600 - 256
				     	else
			 			PDEBUG(1, "irq with packet length = %d and offset=%d for 864", length, vec535->bayeroffset);
					break;
			case	1280:
				 	if (length == 0) 
						vec535->bayeroffset = 0;	//1280 * 864 - 256
				     	else
			 			PDEBUG(1, "irq with packet length = %d and offset=%d for 1280", length, vec535->bayeroffset);
					break;

			case 	2136:
					if (length == 176)  {
			 			PDEBUG(1, "irq with packet length = %d and offset=%d for 2136", length, vec535->bayeroffset);
						vec535->bayeroffset = 3421696;	//2136 * 1602 - 176
					} else {
			 			PDEBUG(1, "irq with packet length = %d and offset=%d for 2136", length, vec535->bayeroffset);
					}
					break;

			case 	2592:
					if (length == 176)  {
			 			PDEBUG(1, "irq with packet length = %d and offset=%d for 2136", length, vec535->bayeroffset);
						vec535->bayeroffset = 3421696;	//2136 * 1602 - 176
					} else {
			 			PDEBUG(1, "irq with packet length = %d and offset=%d for 2136", length, vec535->bayeroffset);
					}
					break;

			default: 	PDEBUG(1, "irq with packet length = %d and offset=%d", length, vec535->bayeroffset);
					return;
		}

	}
	if (length && !urb->status) {
		vec535->nullpackets=0;
		switch(vec535->scratch[vec535->scratch_next].state) {
			case BUFFER_BUSY: {
				vec535->dropped++;
				break;
			}
			case BUFFER_READY: {
				//PDEBUG(3, "Ready and received?! offset=%d", vec535->scratch[vec535->scratch_next].offset);
				vec535->dropped++; //we will take old!!! =)))
				break;
			}
 
			case BUFFER_UNUSED: {
				memcpy(vec535->scratch[vec535->scratch_next].data, (unsigned char *)urb->transfer_buffer, length);
				//if (length < 512) {
				//	memset(vec535->scratch[vec535->scratch_next].data + length, 0, 512 - length);
				//	length = 512;
				//}
				vec535->sts_scratches_free--;
				vec535->sts_scratches_in_use++;
				vec535->scratch[vec535->scratch_next].state=BUFFER_READY;
				vec535->scratch[vec535->scratch_next].offset=vec535->bayeroffset;
				vec535->scratch[vec535->scratch_next].length=length;
				if (waitqueue_active(&vec535->wq)) {
					wake_up_interruptible(&vec535->wq);
				}
				vec535->scratch_next++;
				if (vec535->scratch_next >= VEC535_NUMSCRATCH)
					vec535->scratch_next=0;
				break;
			}
		}
		vec535->bayeroffset+=length;
		if (vec535->bayeroffset>=vec535->cheight*vec535->cwidth) {
			vec535->bayeroffset=0;
		}
	} else {
		vec535->nullpackets++;
		if (vec535->nullpackets > VEC535_MAX_NULLPACKETS) {
			if (waitqueue_active(&vec535->wq)) {
				wake_up_interruptible(&vec535->wq);
			}
		}
	}

	/* Resubmit urb for new data */
	urb->status=0;
	urb->dev=vec535->udev;
	if(usb_submit_urb(urb, GFP_KERNEL))
		PDEBUG(3, "urb burned down");
	return;
}

static int vec535_start_stream(struct usb_vec535 *vec535)
{
	struct urb *urb;
	int err=0, i;
	vec535->streaming=1;
	

	PDEBUG(3, "startstream for %d %d on %d %d mode", vec535->fwidth, vec535->fheight, vec535->cwidth, vec535->cheight);
	for (i=0; i<VEC535_NUMFRAMES; i++) {
		
		vec535->frame[i].data=vec535->fbuf + i * vec535->fwidth * vec535->fheight * 3;
		memset(vec535->frame[i].data, 0x00, vec535->fwidth * vec535->fheight * 3);
		vec535->frame[i].curpix = 0;
	}
	/* Do some memory allocation */
	for (i=0; i<VEC535_NUMSBUF; i++) {
		if ((vec535->sbuf[i].data=kmalloc(VEC535_PACKETSIZE, GFP_KERNEL)) == NULL) {
			err("Can't allocate memory for sttream buffer");
		}
	}
	vec535->bayeroffset=0;
	vec535->scratch_next=0;
	vec535->scratch_use=0;
	for (i=0; i<VEC535_NUMSCRATCH; i++) {
		if ((vec535->scratch[i].data=kmalloc(VEC535_PACKETSIZE, GFP_KERNEL)) == NULL) {
			err("Can't allocate memory for scratch!");
		}
		vec535->scratch[i].state=BUFFER_UNUSED;
	}
	
	PDEBUG(1, "Memory succesfull allocated for buffers");

	/* submitting urbs */
	for (i=0; i<VEC535_NUMSBUF; i++) {
		urb=usb_alloc_urb(0, GFP_KERNEL);
		if(!urb)
			return -ENOMEM;

		usb_fill_bulk_urb(urb, vec535->udev,
			usb_rcvbulkpipe(vec535->udev, VEC535_BULK_IN_ADDR),
			vec535->sbuf[i].data, VEC535_PACKETSIZE,
			vec535_video_irq,
			vec535);

		vec535->urb[i]=urb;

		err=usb_submit_urb(vec535->urb[i], GFP_KERNEL);
		if(err)
			err("urb burned down");
	}
	vec535->framecount=0;
	vec535->sts_scratches_in_use = 0;
	vec535->sts_scratches_free   = VEC535_NUMSCRATCH;

	/* Prepare camera for capturing */
	vec535_control(1, vec535, 0x0A, 0x1F,0x0F);

	return 0;
}

static int vec535_stop_stream(struct usb_vec535 *vec535)
{
	int i;

	if (!vec535->streaming || !vec535->udev)
		return 1;

	vec535->streaming=0;

	/* shutdown camera */
	vec535_control(1, vec535, 0x0A, 0x1D,0x0F);

	//PDEBUG(3, "stop stream for %d %d", vec535->fwidth, vec535->fheight);
	//PDEBUG(1, "STAT ??? Free=%d used=%d dropped=%d", vec535->sts_scratches_free, vec535->sts_scratches_in_use, vec535->dropped);
	for (i=0; i<VEC535_NUMSBUF; i++) 
		if (vec535->urb[i]) {
			usb_kill_urb(vec535->urb[i]);
			usb_free_urb(vec535->urb[i]);
			vec535->urb[i]=NULL;
			kfree(vec535->sbuf[i].data);
		}
	for (i=0; i<VEC535_NUMSCRATCH; i++) {
		kfree(vec535->scratch[i].data);
		vec535->scratch[i].data=NULL;
	}
	PDEBUG(1, "All urb's and buffers was released");

	return 0;
}

/**********************************************************************
 * Color transformation 
 **********************************************************************/

static inline void enhance_picture(unsigned char *frame, int len)
{
	while (len--) {
		*frame=(((*frame^255)*(*frame^255))/255)^255;
		frame++;
	}
}

static int vec535_open(struct file *file)
{		
	struct video_device *dev = video_devdata(file);
	struct usb_vec535 *vec535 = (struct usb_vec535 *)dev;
	int err = 0;

	if (vec535->user)
		return -EBUSY;
	vec535->fbuf = rvmalloc(VEC535_5MPX_SIZE * 3 * VEC535_NUMFRAMES);
	if (vec535->fbuf)
		file->private_data = dev;
	else
		err = -ENOMEM;
	vec535->user = !err;

	return err;
}

static inline void decode_bayer(struct usb_vec535 *vec535, struct vec535_scratch *buffer) {
		int y, x;

	int cwidth = vec535->cwidth;
	int height = vec535->cheight;
	int fwidth = vec535->fwidth;
	int fheight = vec535->fheight;

	int rgbpicsize = fwidth * fheight * 3;
	int pixelR, pixelG, pixelB, pixelY;

	struct vec535_frame *frame=&vec535->frame[vec535->curframe];
	unsigned char *framedata=frame->data;
	

	unsigned char *data=buffer->data;
	int len=buffer->length;
	int i;
	int offset=buffer->offset;
	unsigned char *framedataptr = &framedata[rgbpicsize - offset * 3];
	
/*		 OEOEOEOEOEO
	odd  BGBGBGBGBGB 
    even GRGRGRGRGRG  
	odd  BGBGBGBGBGB 
	even GRGRGRGRGRG 
*/

	//fill the first and last line with black
	if (offset >= cwidth * height) {
		PDEBUG(3, "STOP\n");
	}
	//fill first line with black
	while ((offset < cwidth) && (len)) {
		*(framedataptr - 1) = 0;
		*(framedataptr - 2) = 0;
		*(framedataptr - 3) = 0;
		framedataptr-=3;
		offset++;
		len --;
	}

	
	
	y = offset / cwidth;
	x = offset % cwidth;
	framedataptr = &framedata[rgbpicsize - (y * fwidth + x) * 3];
	
	frame->curpix = framedataptr;
	frame->prevlinepix = framedataptr + fwidth * 3;
		
	//if (y > 870) {
//		PDEBUG(3, "decode bayer on %d %d", x, y);
//	}
	while (len) {

		if ((x >= cwidth)) {
			for (i = cwidth; i < fwidth;i++) {
				*(frame->curpix - 1) = 0;
				*(frame->curpix - 2) = 0;
				*(frame->curpix - 3) = 0;	
				x++;
				frame->curpix -= 3;
				frame->prevlinepix -= 3;

			}
		}else {
			if (y & 1) { 
				if (x & 1) {	//(even, even)
					//BGB
					//GRG
					//BGB
					//FIXME!
					//SETBLUE(offset,  (GETBLUE(offset - cwidth - 1) >> 2) + (GETBLUE(offset - cwidth + 1) >> 2)); 
					*(frame->curpix - 1) = ((*(frame->prevlinepix - 1 - 3)) >> 2) + ((*(frame->prevlinepix - 1 + 3)) >> 2);
					//SETGREEN(offset, (GETGREEN(offset - cwidth) >> 2) + (GETGREEN(offset - 1)>>2)); 
					*(frame->curpix - 2) = ((*(frame->prevlinepix - 2)) >> 2) + ((*(frame->curpix - 2 + 3)) >> 2);
					//SETRED(offset,*data);
					*(frame->curpix - 3) = *data;
				

					//update
					//UPDATERED(offset - cwidth,	((*data) >> 1));
					*(frame->prevlinepix - 3) += ((*data) >> 1); 
				
					//UPDATERED(offset - 1,		((*data) >> 1));
					*(frame->curpix - 3 + 3) += ((*data) >> 1);

					//UPDATERED(offset - 1 - cwidth,((*data) >> 2));
					*(frame->prevlinepix - 3 + 3) += ((*data) >> 2);
				
					//UPDATERED(offset + 1 - cwidth,((*data) >> 2));
					*(frame->prevlinepix - 3 - 3) += ((*data) >> 2);
				}else {			//(even, odd)
					//GBG
					//RGR
					//GBG
					//SETBLUE(offset, GETBLUE(offset - cwidth) >> 1);
					*(frame->curpix - 1) = (*(frame->prevlinepix - 1) >> 1);

					//SETGREEN(offset, *data);
					*(frame->curpix - 2) = *data;

					//SETRED(offset,	GETRED(offset - 1) >> 1);	
					*(frame->curpix - 3) = (*(frame->curpix - 3 + 3)) >> 1;

					//Update
					//UPDATEGREEN(offset - 1,((*data) >> 2));
					*(frame->curpix - 2 + 3) += ((*data) >> 2);
					//UPDATEGREEN(offset - cwidth,((*data) >> 2));
					*(frame->prevlinepix - 2) += ((*data) >> 2);
				}
			}else { 
				if (x & 1) {	//odd, even
					//GRG
					//BGB
					//GRG
					//SETBLUE(offset, GETBLUE(offset - 1) >> 1);	
					*(frame->curpix - 1) = (*(frame->curpix - 1 + 3)) >> 1;
					//SETGREEN(offset, *data);
					*(frame->curpix - 2) = (*data);
					//SETRED(offset,GETRED(offset - cwidth) >> 1);
					*(frame->curpix - 3) = (*(frame->prevlinepix - 3)) >> 1 ;
				
				
					//Update
					//UPDATEGREEN(offset - 1,((*data) >> 2));
					*(frame->curpix - 2 + 3) += ((*data) >> 2);
					//UPDATEGREEN(offset - cwidth,((*data) >> 2));
					*(frame->prevlinepix - 2) += ((*data) >> 2);
				}else {			//odd, odd
					//RGR
					//GBG
					//RGR
					//SETBLUE(offset,*data);
					*(frame->curpix - 1) = *data;
					//SETGREEN(offset, (GETGREEN(offset - cwidth) >> 2) + (GETGREEN(offset - 1) >> 2));
					*(frame->curpix - 2) = ((*(frame->prevlinepix - 2)) >> 2) + ((*(frame->curpix - 2 + 3)) >> 2);
					//SETRED(offset,   (GETRED(offset - cwidth - 1) >> 2) + (GETRED(offset - cwidth + 1)>>2));
					*(frame->curpix - 3) =((*(frame->prevlinepix - 3 - 3)) >> 2) + ((*(frame->prevlinepix - 3 + 3)) >> 2);

					//update 
					//UPDATEBLUE(offset - cwidth	,((*data) >> 1));
					*(frame->prevlinepix - 1) += ((*data) >> 1);
					//UPDATEBLUE(offset - 1		,((*data) >> 1));
					*(frame->curpix - 1 + 3) += ((*data) >> 1);
					//UPDATEBLUE(offset - 1 - cwidth,((*data) >> 2));
					*(frame->prevlinepix - 1 + 3) += ((*data) >> 2);
					//UPDATEBLUE(offset + 1 - cwidth,((*data) >> 2));
					*(frame->prevlinepix - 1 - 3) += ((*data) >> 2);
				}
			}
			//printf("\n");
			len--;
			data++;
			pixelB = *(frame->curpix - 1);
			pixelG = *(frame->curpix - 2);
			pixelR = *(frame->curpix - 3);
 
			pixelY = pixelR * vec535->Sr + pixelG * vec535->Sg + pixelB * vec535->Sb + 128;
			pixelR = (pixelR * vec535->Sx + pixelY) / 256;
			pixelG = (pixelG * vec535->Sx + pixelY) / 256;
			pixelB = (pixelB * vec535->Sx + pixelY) / 256;
		
			if (pixelR > 255) pixelR = 255; if (pixelR < 0) pixelR = 0;
			if (pixelG > 255) pixelG = 255; if (pixelG < 0) pixelG = 0;
			if (pixelB > 255) pixelB = 255; if (pixelB < 0) pixelB = 0;

			*(frame->curpix - 1) = (unsigned char) (pixelB & 0xFF); //BLUE
			*(frame->curpix - 2) = (unsigned char) (pixelG & 0xFF); ; //GREEN
			*(frame->curpix - 3) = (unsigned char) (pixelR & 0xFF); ; //RED
			x++;
			frame->curpix -= 3;
		
			frame->prevlinepix -= 3;
		}


		if (x >= cwidth) {
			for (i = cwidth; i < fwidth;i++) {
				*(frame->curpix - 1) = 0;
				*(frame->curpix - 2) = 0;
				*(frame->curpix - 3) = 0;	
				x++;
				frame->curpix -= 3;
				frame->prevlinepix -= 3;
			}

			y++;
			x = 0;
		}
		if ((y >= height)) {	
			//we should set all the rest to black
			PDEBUG(3, "FOUND EOF x =%d y=%d", x, y);
			//FIXME
                	frame->grabstate=FRAME_DONE;
                	vec535->framecount++;
	                if (vec535->frame[(vec535->curframe+1)&(VEC535_NUMFRAMES-1)].grabstate==FRAME_READY) {
                        	vec535->curframe=(vec535->curframe+1) & (VEC535_NUMFRAMES-1);
                	}
			//if (len)
			//	PDEBUG(3, "FOUND EOF with len=%d (%d, %d)", len, x, y);
			len = 0;

		}

	} 
	

}


/**********************************************************************
 * Building RGB24 frame using scratches  
 **********************************************************************/

static int vec535_newframe(struct usb_vec535 *vec535, int framenr)
{
	DECLARE_WAITQUEUE(wait, current);
	int errors=0;
	int i = 0;

	while (vec535->streaming &&
	    (vec535->frame[framenr].grabstate==FRAME_READY ||
	     vec535->frame[framenr].grabstate==FRAME_GRABBING) ) {
		wait_interruptible(
		    vec535->scratch[vec535->scratch_use].state!=BUFFER_READY,
		    &vec535->wq,
		    &wait
		);
		if (vec535->nullpackets > VEC535_MAX_NULLPACKETS) {
			vec535->nullpackets=0;
			PDEBUG(3, "to many null length packets, restarting capture");
			vec535_stop_stream(vec535);
			vec535_start_stream(vec535);
		} else {
			if (vec535->scratch[vec535->scratch_use].state!=BUFFER_READY) {
				vec535->frame[framenr].grabstate=FRAME_ERROR;
				PDEBUG(1, "Error: No free scratches ??? Free=%d used=%d dropped=%d", vec535->sts_scratches_free, vec535->sts_scratches_in_use, vec535->dropped);
				return -EIO;
			}
			i = 0;
			while(vec535->scratch[vec535->scratch_use].state ==BUFFER_READY) {
				i++;
				vec535->scratch[vec535->scratch_use].state=BUFFER_BUSY;
				decode_bayer(vec535, &vec535->scratch[vec535->scratch_use]);
				vec535->scratch[vec535->scratch_use].state=BUFFER_UNUSED;
				vec535->scratch_use++;
				if (vec535->scratch_use>=VEC535_NUMSCRATCH) {
					vec535->scratch_use=0;
				}
				vec535->sts_scratches_free++;
				vec535->sts_scratches_in_use--;
			}
	//		PDEBUG(1, "%d scr decoded", i);
			if (errors > VEC535_MAX_ERRORS) {
				errors=0;
				PDEBUG(3, "to much errors, restarting capture");
				vec535_stop_stream(vec535);
				vec535_start_stream(vec535);
			}
		}
	}



	//enhance_colour(vec535, vec535->frame[framenr].data, vec535->cheight*vec535->cwidth*3);
	//if ((vec535->frame[framenr].grabstate==FRAME_DONE) &&(enhance)) {
	//		enhance_picture(vec535->frame[framenr].data, vec535->cheight*vec535->cwidth*3);
	//}
	return 0;
}

static int  vec535_getModeNumber(int width, int height) {
	int i;
	for (i = VEC535_MODES_AMMOUNT - 1; i >= 0;i--) {
		if ((width >= width_array[i]) && (height >= height_array[i])) {
			if (disabled_array[i] == 0)
				return i;	
		}
	}
		
	return -1;
}

static long vec535_do_ioctl(struct file *file,
						  			  unsigned int cmd, void *arg)
{
	struct video_device *vdev = file->private_data;
	struct usb_vec535 *vec535 = (struct usb_vec535 *)vdev;
	
	if (!vec535->udev)
		return -EIO;

	switch (cmd) {
	case VIDIOCGCAP: /* Get capabilities */
	{	
		struct video_capability *b = arg;
		PDEBUG(1, "VIDIOCGCAP!");
		strcpy(b->name, VEC535_NAME);
		b->type = VID_TYPE_CAPTURE;
		b->channels = 1;
		b->audios = 0;
		//Only one resolution is supported
		b->maxwidth = width_array[VEC535_MODES_AMMOUNT - 1];	 
		b->maxheight = height_array[VEC535_MODES_AMMOUNT - 1];
		b->minwidth = width_array[0];
		b->minheight = height_array[0];
		return 0;
	}
	case VIDIOCGCHAN: /* Get channels info*/
	{
		struct video_channel *v = arg;
		PDEBUG(1, "VIDIOCGCHAN");

		if (v->channel != 0)
			return -EINVAL;
		v->flags = 0;
		v->tuners = 0;
		v->type = VIDEO_TYPE_CAMERA;
		strcpy(v->name, "Camera");
		return 0;
	}
	case VIDIOCSCHAN: /* Set channel */
	{
		struct video_channel *v = arg;

		PDEBUG(1, "VIDIOCSCHAN!");
		if (v->channel != 0)
			return -EINVAL;
		return 0;
	}
	case VIDIOCGPICT: /* GET Picture properties */
	{
		struct video_picture *p = arg;
		PDEBUG(1, "VIDIOCGPICT!");
		
		p->palette = VIDEO_PALETTE_RGB24;
		p->hue = 32767;    
		p->colour = DEFAULT_SATURATION << 8;
		
		p->brightness = 0; //auto
		p->contrast = 0; //auto
		p->whiteness = 0; //auto
		
		p->depth = 24;
		vec535_get_pict(vec535, p);
		return 0;
	}
	case VIDIOCSPICT: /* SET picture properties */
	{	
		struct video_picture *p = arg;

		vec535_set_pict(vec535, p);
		return 0;
	}
	case VIDIOCSWIN: /* SET the video overlay window */
	{
		struct video_window *vw = arg;
		int mode;
		//return 0;
		PDEBUG(1, "CSWIN:%d, %d, %d, %d", vw->x, vw->y, vw->width, vw->height);
		if (vw->flags)
			return -EINVAL;
		if (vw->clipcount)
			return -EINVAL;
	//	if ((vw->x == 0) && (vw->y == 0) && (vw->width == VEC535_WIDTH) && (vw->height==VEC535_HEIGHT))
	//		return 0;
		
		mode = vec535_getModeNumber(vw->width, vw->height);
		
		if (mode != -1) {	
			if (vec535->streaming)
				vec535_stop_stream(vec535);
				
			vec535_setMode(vec535, mode);
			vec535->fwidth = vw->width;
			vec535->fheight = vw->height;

			if (!vec535->streaming) 
				vec535_start_stream(vec535);

			return 0;
		}

		return -EINVAL;
	}
	case VIDIOCGWIN: /* GET the video overlay window */
	{
		struct video_window *vw = arg;

		vw->x = 0;               /* FIXME */
		vw->y = 0;
		vw->chromakey = 0;
		vw->flags = 0;
		vw->clipcount = 0;
		vw->width = vec535->fwidth;
		vw->height = vec535->fheight;
		return 0;
	}
	case VIDIOCGMBUF: 	/* Memory map buffer info */
	{
		struct video_mbuf *vm = arg;
		int i;
		memset(vm, 0, sizeof(*vm));
		vm->size = VEC535_NUMFRAMES * vec535->fwidth * vec535->fheight * 3;
		vm->frames = VEC535_NUMFRAMES;
		for (i=0; i<VEC535_NUMFRAMES; i++)
			vm->offsets[i] = vec535->fwidth * vec535->fheight * 3 * i;
		return 0;
	}
	case VIDIOCMCAPTURE: /* GRAB FRAMES */
	{
		struct video_mmap *vm = arg;
		int mode, current_mode;
		PDEBUG(1, "VIDIOCMCAPTURE:%d, %d", vm->width, vm->height);
		if (vm->format != VIDEO_PALETTE_RGB24) {
			PDEBUG(3, "ONLY RGB24 is supported, asked for %d", vm->format);
			return -EINVAL;
		}
		if (vm->frame >= VEC535_NUMFRAMES) {
			PDEBUG(3, "too much frames");
			return -EINVAL;
		}
		if (vec535->frame[vm->frame].grabstate != FRAME_UNUSED) {
			PDEBUG(3, "BUSY");
			return -EBUSY;
		}

		/* Is this according to the v4l spec??? */
		//if ((vm->width != VEC535_WIDTH) || (vm->height != VEC535_HEIGHT)) {
		//	PDEBUG(0, "Wrong size (%dx%d)", vm->width, vm->height);
		//	return -EINVAL;
		//}
		mode = vec535_getModeNumber(vm->width, vm->height);
		current_mode = vec535_getModeNumber(vec535->cwidth, vec535->cheight);
		
		if (mode != -1) {	
			PDEBUG(3, "set %dx%d resoulution", vm->width, vm->height);
			vec535->fwidth = vm->width;
			vec535->fheight = vm->height;
			if (current_mode != mode) {
				if (vec535->streaming)
					vec535_stop_stream(vec535);
				
				vec535_setMode(vec535, mode);
			}

			vec535->frame[vm->frame].grabstate=FRAME_READY;
			if (!vec535->streaming) 
				vec535_start_stream(vec535);

			return 0;
		}
		return -EINVAL;

	}
	case VIDIOCSYNC: /* Sync with mmap grabbind */
	{
		int *frame = arg;
		int ret=0;

		//PDEBUG(1, "CSYNC:%d", *frame);
		if(*frame <0 || *frame >= VEC535_NUMFRAMES)
			return -EINVAL;

		ret=vec535_newframe(vec535, *frame);
		vec535->frame[*frame].grabstate=FRAME_UNUSED;
		return ret;
	}
	case VIDIOCGFBUF:	/* GET FRAME BUFFER */
	{
		struct video_buffer *vb = arg;

		memset(vb, 0, sizeof(*vb));
		return 0;
	}
	case VIDIOCKEY:
		return 0;
	case VIDIOCCAPTURE:
		return -EINVAL;
	case VIDIOCSFBUF:
		return -EINVAL;
	case VIDIOCGTUNER:
	case VIDIOCSTUNER:
		return -EINVAL;
	case VIDIOCGFREQ:
	case VIDIOCSFREQ:
		return -EINVAL;
	case VIDIOCGAUDIO:
	case VIDIOCSAUDIO:
		return -EINVAL;
	default:
		return -ENOIOCTLCMD;
	} /* end switch */
	return 0;
}

static void usb_vec535_remove_disconnected (struct usb_vec535 *vec535)
{
	int i;

	vec535->udev = NULL;

	for (i=0; i<VEC535_NUMSBUF; i++)
		if (vec535->urb[i]) {
			usb_kill_urb(vec535->urb[i]);
			usb_free_urb(vec535->urb[i]);
			vec535->urb[i] = NULL;
			kfree(vec535->sbuf[i].data);
		}
	for (i=0; i<VEC535_NUMSCRATCH; i++) {
		kfree(vec535->scratch[i].data);
	}
	PDEBUG(1, "%s disconnected", VEC535_NAME);

	/* Free the memory */
	kfree(vec535);
}

static int vec535_close(struct file *file)
{
	struct video_device *dev = file->private_data;
	struct usb_vec535 *vec535 = (struct usb_vec535 *)dev;
	int i;

	rvfree(vec535->fbuf, VEC535_5MPX_SIZE * 3 * VEC535_NUMFRAMES );
	if (vec535->removed) {
		usb_vec535_remove_disconnected(vec535);
		PDEBUG(1, "device unregistered");
	} else {
		for (i=0; i<VEC535_NUMFRAMES; i++)
			vec535->frame[i].grabstate=FRAME_UNUSED;
		if (vec535->streaming)
			vec535_stop_stream(vec535);
		vec535->user=0;
	}
	file->private_data = NULL;
	return 0;
}
static long vec535_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
	return video_usercopy(file, cmd, arg, vec535_do_ioctl);
}

static ssize_t vec535_read(struct file *file, char __user *buf, size_t count, loff_t *ppos)
{
	int realcount=count, ret=0;
	struct video_device *dev = file->private_data;
	struct usb_vec535 *vec535 = (struct usb_vec535 *)dev;
	
	PDEBUG(3, "vec535_read!");

	if (vec535->udev == NULL)
		return -EIO;

	//output is RGB24 !!!
	if (realcount > vec535->fwidth*vec535->fheight * 3)
		realcount=vec535->fwidth*vec535->fheight * 3;

	if (vec535->frame[0].grabstate==FRAME_GRABBING) {
		PDEBUG(1, "grabstate == FRAME_GRABBING??");
		return -EBUSY;
	}
	
	vec535->frame[0].grabstate=FRAME_READY;
	vec535->frame[1].grabstate=FRAME_UNUSED;
	vec535->curframe=0;

	if (!vec535->streaming)
		vec535_start_stream(vec535);

	ret=vec535_newframe(vec535, 0);


	vec535->frame[0].grabstate=FRAME_UNUSED;
	if (ret)
		return ret;

	if (copy_to_user(buf, vec535->frame[0].data, realcount)) {
		return -EFAULT;
	}

	return realcount;

}
static int vec535_mmap(struct file *file, struct vm_area_struct *vma)
{
	struct video_device *dev = file->private_data;
	struct usb_vec535 *vec535 = (struct usb_vec535 *)dev;
	unsigned long start = vma->vm_start;
	unsigned long size  = vma->vm_end-vma->vm_start;
	unsigned long page, pos;

	mutex_lock(&vec535->lock);

	if (vec535->udev == NULL) {
		mutex_unlock(&vec535->lock);
		return -EIO;
	}
	if (size > (((VEC535_NUMFRAMES * vec535->fwidth * vec535->fheight * 3) + PAGE_SIZE - 1) & ~(PAGE_SIZE - 1))) {
		mutex_unlock(&vec535->lock);
		return -EINVAL;
	}
	pos = (unsigned long)vec535->fbuf;
	while (size > 0) {
		page = vmalloc_to_pfn((void *)pos);
		if (remap_pfn_range(vma, start, page, PAGE_SIZE, PAGE_SHARED)) {
			mutex_unlock(&vec535->lock);
			return -EAGAIN;
		}
		start += PAGE_SIZE;
		pos += PAGE_SIZE;
		if (size > PAGE_SIZE)
			size -= PAGE_SIZE;
		else
			size = 0;
	}
	mutex_unlock(&vec535->lock);

	return 0;
}

static const struct v4l2_file_operations vec535_fops = {
	.owner =	THIS_MODULE,
	.open =         vec535_open,
	.release =      vec535_close,
	.read =         vec535_read,
	.mmap =         vec535_mmap,
	.ioctl =        vec535_ioctl,
};
static struct video_device vec535_template = {
	.name =         VEC535_NAME,
	.fops =         &vec535_fops,
	.release = video_device_release_empty,
};


static int vec535_probe(struct usb_interface *interface,	const struct usb_device_id *id)
{
	struct usb_device *dev = interface_to_usbdev(interface);
	struct usb_host_interface *h_interface = interface->cur_altsetting;
	struct usb_vec535 *vec535; 


	/* We don't handle multi-config cameras */
	if (dev->descriptor.bNumConfigurations != 1)
		return -ENODEV;

	/* Is it an vec535? */
	if ((le16_to_cpu(dev->descriptor.idVendor) == 0x04b4) &&
	    (le16_to_cpu(dev->descriptor.idProduct) == 0x0535) && 
		(h_interface->desc.bInterfaceClass == 0xFF) &&
		(h_interface->desc.bInterfaceSubClass == 0x00)) {
		PDEBUG(0, "EVS VEC 535 camera found");
	}else 		
		return -ENODEV;
	
	//Alloc memory for vec535 structure
	if ((vec535 = kzalloc(sizeof(struct usb_vec535), GFP_KERNEL)) == NULL) {
		err("couldn't kmalloc vec535 struct");
		return -ENOMEM;
	}
	
	vec535->udev = dev;
	vec535->interface = interface;

	vec535_init(vec535);
	//Setting default resolution
	vec535_setMode(vec535, MODE_2136x1602);
	vec535->fwidth = 2136;
	vec535->fheight = 1602;

	memcpy(&vec535->vdev, &vec535_template, sizeof(vec535_template));
	memcpy(vec535->vdev.name, VEC535_NAME, strlen(VEC535_NAME));
	init_waitqueue_head(&vec535->wq);
	mutex_init(&vec535->lock);
	wmb();

	if (video_register_device(&vec535->vdev, VFL_TYPE_GRABBER, video_nr) == -1) {
		kfree(vec535);
		err("video_register_device failed");
		return -EIO;
	}								
	PDEBUG(1, "registered new video device: video%d", vec535->vdev.minor);

	usb_set_intfdata (interface, vec535);
	return 0;
}

static void vec535_disconnect(struct usb_interface *intf)
{
	struct usb_vec535 *vec535 = usb_get_intfdata (intf);

	usb_set_intfdata (intf, NULL);
	if (vec535) {
		video_unregister_device(&vec535->vdev);
		if (!vec535->user){
			usb_vec535_remove_disconnected(vec535);
		} else {
			vec535->frame[0].grabstate = FRAME_ERROR;
			vec535->frame[1].grabstate = FRAME_ERROR;

			vec535->streaming = 0;

			wake_up_interruptible(&vec535->wq);
			vec535->removed = 1;
		}
	}
}

static struct usb_driver vec535_driver = {
	.name = "vec535",
	.id_table = device_table,
	.probe = vec535_probe,
	.disconnect = vec535_disconnect
};

static int __init vec535_init_module(void)
{
	PDEBUG(0, "Module vec535 init (version %s)", version);
	return usb_register(&vec535_driver);
}

static void __exit vec535_exit_module(void)
{
	usb_deregister(&vec535_driver);
	PDEBUG(0, "Module vec535 exit" );
}

module_init(vec535_init_module);
module_exit(vec535_exit_module);
